from mvesuvio.analysis_fitting import FitInYSpace
from mvesuvio.util import handle_config
from mvesuvio.util.analysis_helpers import (
    calculate_resolution,
    fix_profile_parameters,
    isolate_lighest_mass_data,
    load_raw_and_empty_from_path,
    cropAndMaskWorkspace,
    calculate_h_ratio,
    ws_history_matches_inputs,
    save_ws_from_load_vesuvio,
    is_hydrogen_present,
    create_profiles_table,
    create_table_for_hydrogen_to_mass_ratios,
    print_table_workspace,
    convert_to_list_of_spectrum_numbers,
)
from mvesuvio.analysis_reduction import VesuvioAnalysisRoutine
from mvesuvio import globals

from mantid.api import AnalysisDataService
from mantid.simpleapi import mtd, RenameWorkspace, SaveAscii, Load
from mantid.kernel import logger
from mantid.api import AlgorithmFactory, AlgorithmManager

import matplotlib.pyplot as plt
import numpy as np
from pathlib import Path
import importlib
import sys
import dill  # To convert constraints to string
import re
import os
import traceback


class Runner:
    def __init__(
        self,
        override_back_workspace="",
        override_front_workspace="",
        bootstrap_inputs_directory="",
        minimal_output=False,
        output_directory="",
        running_tests=False,
    ) -> None:
        self.running_tests = running_tests
        self.inputs_path = Path(handle_config.read_cached_var("caching.inputs"))
        self.override_back_workspace = override_back_workspace
        self.override_front_workspace = override_front_workspace
        self.bootstrap_inputs_directory = bootstrap_inputs_directory
        self.minimal_output = minimal_output
        self.mantid_log_file = "mantid.log"

        # I/O paths
        inputs_script_path = Path(handle_config.read_cached_var("caching.inputs"))
        script_name = handle_config.get_script_name()
        self.experiment_path = inputs_script_path.parent / script_name
        self.input_ws_path = self.experiment_path / "input_workspaces"
        self.input_ws_path.mkdir(parents=True, exist_ok=True)
        if not output_directory:
            self.output_directory = Path(self.experiment_path) / "output_files"
        else:
            self.output_directory = Path(output_directory)
        self.reduction_directory = self.output_directory / "reduction"
        self.fitting_directory = self.output_directory / "fitting"

        self.analysis_result = None
        self.fitting_result = None

        self.setup()

    def setup(self):
        ai = self.import_from_inputs()

        self.bckwd_ai = ai.BackwardAnalysisInputs
        self.fwd_ai = ai.ForwardAnalysisInputs

        self.fwd_ai.minimal_output = self.minimal_output
        self.bckwd_ai.minimal_output = self.minimal_output

        self.fwd_ai.name = handle_config.get_script_name() + "_" + globals.FORWARD_TAG
        self.bckwd_ai.name = handle_config.get_script_name() + "_" + globals.BACKWARD_TAG

        self.fwd_ai.override_input_workspace = self.override_front_workspace
        self.bckwd_ai.override_input_workspace = self.override_back_workspace

        self.update_ws_names_from_override_input_workspaces()

    def import_from_inputs(self):
        name = "analysis_inputs"
        spec = importlib.util.spec_from_file_location(name, self.inputs_path)
        module = importlib.util.module_from_spec(spec)
        sys.modules[name] = module
        spec.loader.exec_module(module)
        return module

    def run_bootstrap(self):
        if self.h_ratio_is_zero_when_h_present():
            logger.error("Hydrogen ratio not set, run analysis on sample first before attempting bootstrap.")
            return

        input_dirs = self.get_bootstrap_input_directories()
        if not input_dirs:
            return
        inputs_parent_path, inputs_backward_path, inputs_forward_path = input_dirs

        # Sort files based on last character, assumed to be integer
        def sorting_order(p):
            return int(p.stem[-1])

        boot_outputs_dir_path = inputs_parent_path.parent / (inputs_parent_path.name + "_outputs")
        boot_outputs_dir_path.mkdir(exist_ok=True)

        if self.bckwd_ai.run_this_scattering_type and self.fwd_ai.run_this_scattering_type:
            procedure_output_dir_path = boot_outputs_dir_path / "joint"
            procedure_output_dir_path.mkdir(exist_ok=True)

            for back_ws_path, front_ws_path in zip(
                sorted(inputs_backward_path.iterdir(), key=sorting_order), sorted(inputs_forward_path.iterdir(), key=sorting_order)
            ):
                common_prefix = self.get_common_prefix_of_bootstrap_sample_names(back_ws_path.stem, front_ws_path.stem)
                if not common_prefix:
                    return
                sample_output_directory = procedure_output_dir_path / (common_prefix + "_joint_" + back_ws_path.stem[-1])

                self.update_sample_inputs_outputs(back_ws_path, front_ws_path, sample_output_directory)
                AnalysisDataService.clear()
                self.run()
                plt.close("all")  # Close any plots that might still be open
            return

        if self.bckwd_ai.run_this_scattering_type:
            procedure_output_dir_path = boot_outputs_dir_path / "backward"
            procedure_output_dir_path.mkdir(exist_ok=True)
            for back_ws_path in sorted(inputs_backward_path.iterdir(), key=sorting_order):
                sample_output_directory = procedure_output_dir_path / (back_ws_path.stem.split("_")[0] + "_bckwd_" + back_ws_path.stem[-1])

                self.update_sample_inputs_outputs(back_ws_path, None, sample_output_directory)
                AnalysisDataService.clear()
                self.run()
                plt.close("all")  # Close any plots that might still be open
            return

        if self.fwd_ai.run_this_scattering_type:
            procedure_output_dir_path = boot_outputs_dir_path / "forward"
            procedure_output_dir_path.mkdir(exist_ok=True)
            for front_ws_path in sorted(inputs_forward_path.iterdir(), key=sorting_order):
                sample_output_directory = procedure_output_dir_path / (front_ws_path.stem.split("_")[0] + "_fwd_" + front_ws_path.stem[-1])

                self.update_sample_inputs_outputs(None, front_ws_path, sample_output_directory)
                AnalysisDataService.clear()
                self.run()
                plt.close("all")  # Close any plots that might still be open
            return
        return

    def update_sample_inputs_outputs(self, back_ws_path, front_ws_path, output_path):
        output_path.mkdir(exist_ok=True)
        self.update_output_directory(output_path)

        self.bckwd_ai.override_input_workspace = str(back_ws_path.absolute()) if back_ws_path is not None else ""
        self.fwd_ai.override_input_workspace = str(front_ws_path.absolute()) if front_ws_path is not None else ""
        self.update_ws_names_from_override_input_workspaces()

        self.bckwd_ai.show_plots = False
        self.fwd_ai.show_plots = False
        return

    def get_common_prefix_of_bootstrap_sample_names(self, back_ws_name, front_ws_name):
        if back_ws_name == front_ws_name:
            logger.error(f"Bootstrap Error: backward and forward inputs should not have the same name: {front_ws_name}.")
            return ""
        if back_ws_name[-1] != front_ws_name[-1]:
            logger.error(f"Bootstrap Error: inputs {back_ws_name} and {front_ws_name} do not have the same last character.")
            return ""

        def longest_common_prefix(s1, s2):
            return s1[: next((i for i, (a, b) in enumerate(zip(s1, s2)) if a != b or a == "_" or b == "_"), min(len(s1), len(s2)))]

        common_prefix = longest_common_prefix(back_ws_name, front_ws_name)
        if not common_prefix:
            logger.error(f"Bootstrap Error: inputs {back_ws_name} and {front_ws_name} do not have a common prefix.")
            return ""
        return common_prefix

    def get_bootstrap_input_directories(self):
        inputs_dir_path = Path(self.bootstrap_inputs_directory)
        if not inputs_dir_path.is_dir():
            logger.error("The inputs directory path provided for bootstrap is not a directory.")
            return ()

        inputs_backward_path = inputs_dir_path / "backward"
        if not inputs_backward_path.exists():
            inputs_backward_path.mkdir(exist_ok=True)
            logger.error(f"Created backward directory. Please place your backward samples here: {str(inputs_backward_path)}")
            return ()

        inputs_forward_path = inputs_dir_path / "forward"
        if not inputs_forward_path.exists():
            inputs_forward_path.mkdir(exist_ok=True)
            logger.error(f"Created forward directory. Please place your forward samples here: {str(inputs_forward_path)}")
            return ()
        return (inputs_dir_path, inputs_backward_path, inputs_forward_path)

    def update_output_directory(self, path):
        self.output_directory = path
        self.reduction_directory = self.output_directory / "reduction"
        self.fitting_directory = self.output_directory / "fitting"

    def update_ws_names_from_override_input_workspaces(self):
        if self.fwd_ai.override_input_workspace:
            self.fwd_ai.name = Path(self.fwd_ai.override_input_workspace).stem

        if self.bckwd_ai.override_input_workspace:
            self.bckwd_ai.name = Path(self.bckwd_ai.override_input_workspace).stem

    def run(self):
        if not self.bckwd_ai.run_this_scattering_type and not self.fwd_ai.run_this_scattering_type:
            return

        # Erase previous log
        # Not working on Windows due to shared file locks
        if os.name == "posix":
            with open(self.mantid_log_file, "w") as file:
                file.write("")

        if self.runAnalysisFitting():
            self.make_summarised_log_file()
            return self.analysis_result, self.fitting_result

        if self.h_ratio_is_zero_when_h_present():
            self.run_estimate_h_ratio()
            self.make_summarised_log_file()
            return

        self.runAnalysisRoutine()
        self.runAnalysisFitting()

        self.make_summarised_log_file()
        # Return results used only in tests
        return self.analysis_result, self.fitting_result

    def h_ratio_is_zero_when_h_present(self):
        if self.bckwd_ai.run_this_scattering_type:
            if is_hydrogen_present(self.fwd_ai.masses):
                if self.bckwd_ai.intensity_ratio_of_hydrogen_to_chosen_mass == 0:
                    return True
            else:
                logger.warning("Ignoring Hydrogen ratio because not detected in masses.")
                self.bckwd_ai.intensity_ratio_of_hydrogen_to_chosen_mass = 0
        return False

    def make_summarised_log_file(self):
        pattern = re.compile(r"^\d{4}-\d{2}-\d{2}")

        log_file_save_path = self.make_log_file_name()

        try:
            with open(self.mantid_log_file, "r") as infile, open(log_file_save_path, "w") as outfile:
                for line in infile:
                    if "VesuvioAnalysisRoutine" in line:
                        outfile.write(line)

                    if "Notice Python" in line:  # For Fitting notices
                        outfile.write(line)

                    if not pattern.match(line):
                        outfile.write(line)
        except OSError:
            logger.error("Mantid log file not available. Unable to produce a summarized log file for this routine.")
        return

    def make_log_file_name(self):
        filename = ""
        if self.bckwd_ai.run_this_scattering_type:
            filename += "bckwd_" + self.bckwd_ai.fitting_model
        if self.fwd_ai.run_this_scattering_type:
            filename += "fwd_" + self.fwd_ai.fitting_model
        return self.experiment_path / (filename + ".log")

    def runAnalysisFitting(self):
        success = False
        if self.bckwd_ai.fit_in_y_space:
            try:
                success |= self.run_y_space_reduction_and_fit(self.bckwd_ai)
            except Exception:
                success |= True
                logger.error(traceback.format_exc())
        if self.fwd_ai.fit_in_y_space:
            try:
                success |= self.run_y_space_reduction_and_fit(self.fwd_ai)
            except Exception:
                success |= True
                logger.error(traceback.format_exc())
        return success

    def run_y_space_reduction_and_fit(self, ai):
        ws_to_fit_name = ai.name + "_" + str(ai.number_of_iterations_for_corrections)
        try:
            ws_to_fit = mtd[ws_to_fit_name]
            ws_to_fit_ncps = mtd[ws_to_fit_name + "_ncp_group"]
        except KeyError:
            logger.warning(f"Either {ws_to_fit_name} or {ws_to_fit_name}_ncp_group not found. Skipping fitting in Y-Space.")
            return False

        # TODO: Move resolution calculation to end of analysis, instead of beggining of fitting
        ws_resolution = calculate_resolution(min(ai.masses), ws_to_fit, ai.range_for_rebinning_in_y_space)
        # NOTE: Resolution workspace is useful for scientists outside mantid
        self.fitting_directory.mkdir(exist_ok=True)
        SaveAscii(ws_resolution.name(), str(self.fitting_directory / ws_resolution.name()))

        ws_lighest_data, ws_lighest_ncp = isolate_lighest_mass_data(ws_to_fit, ws_to_fit_ncps, ai.subtract_calculated_fse_from_data)
        self.fitting_result = FitInYSpace(ai, ws_lighest_data, ws_lighest_ncp, ws_resolution, outputs_dir=self.fitting_directory).run()
        return True

    def runAnalysisRoutine(self):
        if self.bckwd_ai.override_input_workspace:
            if not self.bckwd_ai.override_input_workspace.endswith(".nxs"):
                logger.error("Currently only nexus files are supported for overriding input workspaces. Try again with a .nxs file.")
                return
        if self.fwd_ai.override_input_workspace:
            if not self.fwd_ai.override_input_workspace.endswith(".nxs"):
                logger.error("Currently only nexus files are supported for overriding input workspaces. Try again with a .nxs file.")
                return

        if self.bckwd_ai.run_this_scattering_type and self.fwd_ai.run_this_scattering_type:
            # Either both overrides are empty or they're both set to a workspace path
            if (not self.bckwd_ai.override_input_workspace) != (not self.fwd_ai.override_input_workspace):
                logger.error(
                    "Running joint analysis requires that both backward and forward workspaces are explicitly passed as inputs to override default input workspaces."
                )
                return
            self.run_joint_analysis()
            return

        if self.bckwd_ai.run_this_scattering_type:
            if not self.bckwd_ai.override_input_workspace and self.fwd_ai.override_input_workspace:
                logger.error(
                    "Forward input workspace was explicitly set but analysis is running backward routine. Please provide input workspace for backward analysis."
                )
                return

            self.run_single_analysis(self.bckwd_ai)
            return

        if self.fwd_ai.run_this_scattering_type:
            if not self.fwd_ai.override_input_workspace and self.bckwd_ai.override_input_workspace:
                logger.error(
                    "Backward input workspace was explicitly set but analysis is running forward routine. Please provide input workspace for forward analysis."
                )
                return
            self.run_single_analysis(self.fwd_ai)
            return
        return

    def run_single_analysis(self, ai):
        AnalysisDataService.clear()
        alg = self._create_analysis_algorithm(ai)
        alg.execute()
        self.analysis_result = alg
        return

    def run_joint_analysis(self):
        AnalysisDataService.clear()
        back_alg = self._create_analysis_algorithm(self.bckwd_ai)
        front_alg = self._create_analysis_algorithm(self.fwd_ai)
        self.run_joint_algs(back_alg, front_alg)
        return

    @classmethod
    def run_joint_algs(cls, back_alg, front_alg):
        back_alg.execute()

        incoming_means_table = mtd[back_alg.getPropertyValue("OutputMeansTable")]
        h_ratio = back_alg.getProperty("HRatioToChosenMass").value

        assert incoming_means_table is not None, "Means table from backward routine not correctly accessed."
        assert h_ratio is not None, "H ratio from backward routine not correctly accesssed."

        receiving_profiles_table = mtd[front_alg.getPropertyValue("InputProfiles")]

        fixed_profiles_table = fix_profile_parameters(incoming_means_table, receiving_profiles_table, h_ratio)

        # Update original profiles table
        RenameWorkspace(fixed_profiles_table, receiving_profiles_table.name())
        print_table_workspace(mtd[receiving_profiles_table.name()])
        # Even if the name is the same, need to trigger update
        front_alg.setPropertyValue("InputProfiles", receiving_profiles_table.name())

        front_alg.execute()
        return

    def run_estimate_h_ratio(self):
        """
        Used when H is present and H to first mass ratio is not known.
        Preliminary forward scattering is run to get rough estimate of H to first mass ratio.
        Runs iterative procedure with alternating back and forward scattering.
        """

        if not self.running_tests:
            try:
                userInput = input("\nHydrogen intensity ratio to lowest mass is not set. Press Enter to start estimate procedure.")
                if not userInput == "":
                    raise EOFError

            except EOFError:
                logger.error("Estimation of Hydrogen intensity ratio interrupted.")
                return

        chosen_mass = self.bckwd_ai.masses[self.bckwd_ai.chosen_mass_index]

        table_h_ratios = create_table_for_hydrogen_to_mass_ratios()

        back_alg = self._create_analysis_algorithm(self.bckwd_ai)
        front_alg = self._create_analysis_algorithm(self.fwd_ai)

        front_alg.execute()

        means_table = mtd[front_alg.getPropertyValue("OutputMeansTable")]
        current_ratio = calculate_h_ratio(means_table, chosen_mass)

        table_h_ratios.addRow([current_ratio])
        previous_ratio = np.nan

        while not np.isclose(current_ratio, previous_ratio, rtol=0.01):
            back_alg.setProperty("HRatioToChosenMass", current_ratio)
            self.run_joint_algs(back_alg, front_alg)

            previous_ratio = current_ratio

            means_table = mtd[front_alg.getPropertyValue("OutputMeansTable")]
            current_ratio = calculate_h_ratio(means_table, chosen_mass)

            table_h_ratios.addRow([current_ratio])

            SaveAscii(table_h_ratios.name(), str(self.output_directory / table_h_ratios.name()))

        logger.notice("\nProcedute to estimate Hydrogen ratio finished.")
        print_table_workspace(table_h_ratios)
        return

    def _create_analysis_algorithm(self, ai):
        if not ai.override_input_workspace:
            raw_path, empty_path = self._save_ws_if_not_on_path(ai)
            ws = load_raw_and_empty_from_path(
                raw_path=raw_path,
                empty_path=empty_path,
                tof_binning=ai.time_of_flight_binning,
                name=ai.name,
                raw_scale_factor=ai.scale_raw_workspace,
                empty_scale_factor=ai.scale_empty_workspace,
                raw_minus_empty=ai.subtract_empty_workspace_from_raw,
            )
            first_detector, last_detector = [int(s) for s in ai.detectors.split("-")]
            input_ws = cropAndMaskWorkspace(
                ws,
                firstSpec=first_detector,
                lastSpec=last_detector,
                maskedDetectors=ai.mask_detectors,
                maskTOFRange=ai.mask_time_of_flight_range,
            )
        else:
            input_ws = Load(Filename=str(Path(ai.override_input_workspace).absolute()), OutputWorkspace=ai.name)

        profiles_table = create_profiles_table(input_ws.name() + "_initial_parameters", ai)
        print_table_workspace(profiles_table)
        ipFilesPath = Path(handle_config.read_cached_var("caching.ipfolder"))
        kwargs = {
            "InputWorkspace": input_ws.name(),
            "InputProfiles": profiles_table.name(),
            "InstrumentParametersFile": str(ipFilesPath / ai.instrument_parameters_file),
            "ChosenMassIndex": ai.chosen_mass_index if hasattr(ai, "chosen_mass_index") else 0,
            "HRatioToChosenMass": ai.intensity_ratio_of_hydrogen_to_chosen_mass
            if hasattr(ai, "intensity_ratio_of_hydrogen_to_chosen_mass")
            else 0,
            "NumberOfIterations": int(ai.number_of_iterations_for_corrections),
            "InvalidDetectors": convert_to_list_of_spectrum_numbers(ai.mask_detectors),
            "MultipleScatteringCorrection": ai.do_multiple_scattering_correction,
            "SampleShapeXml": ai.sample_shape_xml,
            "GammaCorrection": ai.do_gamma_correction,
            "ModeRunning": "BACKWARD" if self.get_scattering_type(ai) == globals.BACKWARD_TAG else "FORWARD",
            "TransmissionGuess": ai.transmission_guess,
            "MultipleScatteringOrder": int(ai.multiple_scattering_order),
            "NumberOfEvents": int(ai.multiple_scattering_number_of_events),
            "Constraints": str(dill.dumps(ai.constraints)),
            "ResultsPath": str(self.reduction_directory.absolute()),
            "MinimalOutputFiles": ai.minimal_output,
            "OutputMeansTable": " Final_Means",
        }

        if self.running_tests:
            alg = VesuvioAnalysisRoutine()
        else:
            AlgorithmFactory.subscribe(VesuvioAnalysisRoutine)
            alg = AlgorithmManager.createUnmanaged("VesuvioAnalysisRoutine")

        alg.initialize()
        alg.setProperties(kwargs)
        return alg

    def _save_ws_if_not_on_path(self, ai):
        raw_name = handle_config.get_script_name() + "_" + "raw" + "_" + self.get_scattering_type(ai) + ".nxs"
        empty_name = handle_config.get_script_name() + "_" + "empty" + "_" + self.get_scattering_type(ai) + ".nxs"

        raw_path = self.input_ws_path / raw_name
        empty_path = self.input_ws_path / empty_name

        ip_files_path = Path(handle_config.read_cached_var("caching.ipfolder"))

        if not ws_history_matches_inputs(ai.runs, ai.mode, ai.instrument_parameters_file, raw_path):
            save_ws_from_load_vesuvio(ai.runs, ai.mode, str(ip_files_path / ai.instrument_parameters_file), raw_path)

        if not ws_history_matches_inputs(ai.empty_runs, ai.mode, ai.instrument_parameters_file, empty_path):
            save_ws_from_load_vesuvio(ai.empty_runs, ai.mode, str(ip_files_path / ai.instrument_parameters_file), empty_path)
        return raw_path, empty_path

    def get_scattering_type(self, ai):
        if ai.__name__ in ["BackwardAnalysisInputs"]:
            return globals.BACKWARD_TAG
        elif ai.__name__ in ["ForwardAnalysisInputs"]:
            return globals.FORWARD_TAG
        else:
            raise ValueError(f"Input class for workspace not valid: {ai.__name__}")


if __name__ == "__main__":
    Runner().run()



"""
Vesuvio
=============

Vesuvio is an instrument that performs Neuton Compton Scattering, based at ISIS, RAL, UK. This code processes raw output data to determine
nuclear kinetic energies and moment distributions.
"""

from mvesuvio._version import __version__
from mvesuvio.main import main

__project_url__ = "https://github.com/mantidproject/vesuvio"

__all__ = ["__version__", "__project_url__"]


class ArgInputs:
    def __init__(self, command):
        self.__command = command

    @property
    def command(self):
        return self.__command


class ConfigArgInputs(ArgInputs):
    def __init__(self, analysis_inputs, ip_folder):
        super().__init__("config")
        self.__set_inputs = analysis_inputs
        self.__set_ipfolder = ip_folder

    @property
    def analysis_inputs(self):
        return self.__set_inputs

    @property
    def ip_folder(self):
        return self.__set_ipfolder


class RunArgInputs(ArgInputs):
    def __init__(self, back_workspace, front_workspace, minimal_output, outputs_dir):
        super().__init__("run")
        self.__back_workspace = back_workspace
        self.__front_workspace = front_workspace
        self.__minimal_output = minimal_output
        self.__outputs_dir = outputs_dir

    @property
    def back_workspace(self):
        return self.__back_workspace

    @property
    def front_workspace(self):
        return self.__front_workspace

    @property
    def minimal_output(self):
        return self.__minimal_output

    @property
    def outputs_dir(self):
        return self.__outputs_dir


def config(analysis_inputs="", ip_folder=""):
    config_args = ConfigArgInputs(analysis_inputs, ip_folder)
    main(config_args)


def run(back_workspace="", front_workspace="", minimal_output=False, outputs_dir=""):
    run_args = RunArgInputs(back_workspace, front_workspace, minimal_output, outputs_dir)
    main(run_args)


def version():
    """Returns the version of the mvesuvio package."""
    return __version__



from mantid.simpleapi import (
    Load,
    Rebin,
    Scale,
    SumSpectra,
    Minus,
    CropWorkspace,
    MaskDetectors,
    CreateEmptyTableWorkspace,
    DeleteWorkspace,
    SaveNexus,
    LoadVesuvio,
    mtd,
    VesuvioResolution,
    AppendSpectra,
    RenameWorkspace,
    CloneWorkspace,
)
from mantid.kernel import logger
import numpy as np

from mvesuvio import globals

import ntpath


def isolate_lighest_mass_data(initial_ws, ws_group_ncp, subtract_fse=True):
    # NOTE: Minus() is not used so it doesn't change dataE

    ws_ncp_names = [n for n in ws_group_ncp.getNames() if n.endswith("ncp")]
    masses = [float(n.split("_")[-2]) for n in ws_ncp_names if "total" not in n]
    ws_total_ncp_name = [n for n in ws_ncp_names if n.endswith("_total_ncp")][0]
    ws_lighest_ncp_name = ws_ncp_names[masses.index(min(masses))]
    ws_lighest_ncp = mtd[ws_lighest_ncp_name]
    ws_total_ncp = mtd[ws_total_ncp_name]
    suffix = "_m0"

    # Main subtraction
    isolated_data_y = initial_ws.extractY() - (ws_total_ncp.extractY() - ws_lighest_ncp.extractY())

    if subtract_fse:
        suffix += "_-fse"
        ws_lighest_fse = mtd[ws_lighest_ncp_name.replace("ncp", "fse")]

        isolated_data_y -= ws_lighest_fse.extractY()

        # Subtract from fitted ncp
        # TODO: Find a better solution later
        ws_lighest_ncp_y = ws_lighest_ncp.extractY()
        ws_lighest_ncp_y -= ws_lighest_fse.extractY()
        ws_lighest_ncp = CloneWorkspace(ws_lighest_ncp, OutputWorkspace=ws_lighest_ncp.name() + "_-fse")
        write_data_y_into_ws(ws_lighest_ncp_y, ws_lighest_ncp)
        SumSpectra(ws_lighest_ncp.name(), OutputWorkspace=ws_lighest_ncp.name() + "_sum")

    # TODO: Find a better way to propagate masked values
    isolated_data_y[initial_ws.extractY() == 0] = 0
    ws_lighest_data = CloneWorkspace(initial_ws, OutputWorkspace=initial_ws.name() + suffix)
    write_data_y_into_ws(isolated_data_y, ws_lighest_data)
    SumSpectra(ws_lighest_data.name(), OutputWorkspace=ws_lighest_data.name() + "_sum")

    return ws_lighest_data, ws_lighest_ncp


def write_data_y_into_ws(data_y, ws):
    for i in range(ws.getNumberHistograms()):
        ws.dataY(i)[:] = data_y[i, :]
    return


def calculate_resolution(mass, ws, rebin_range):
    resName = ws.name() + "_resolution"
    for index in range(ws.getNumberHistograms()):
        VesuvioResolution(Workspace=ws, WorkspaceIndex=index, Mass=mass, OutputWorkspaceYSpace="tmp")
        Rebin(
            InputWorkspace="tmp",
            Params=rebin_range,
            OutputWorkspace="tmp",
        )

        if index == 0:  # Ensures that workspace has desired units
            RenameWorkspace("tmp", resName)
        else:
            AppendSpectra(resName, "tmp", OutputWorkspace=resName)

    masked_idx = [ws.spectrumInfo().isMasked(i) for i in range(ws.getNumberHistograms())]
    MaskDetectors(resName, WorkspaceIndexList=np.flatnonzero(masked_idx))
    DeleteWorkspace("tmp")
    return mtd[resName]


def pass_data_into_ws(dataX, dataY, dataE, ws):
    "Modifies ws data to input data"
    for i in range(ws.getNumberHistograms()):
        ws.dataX(i)[:] = dataX[i, :]
        ws.dataY(i)[:] = dataY[i, :]
        ws.dataE(i)[:] = dataE[i, :]
    return ws


def print_table_workspace(table, precision=3):
    table_dict = table.toDict()
    # Convert floats into strings
    for key, values in table_dict.items():
        new_column = [int(item) if (isinstance(item, float) and item.is_integer()) else item for item in values]
        table_dict[key] = [f"{item:.{precision}f}" if isinstance(item, float) else str(item) for item in new_column]

    max_spacing = [max([len(item) for item in values] + [len(key)]) for key, values in table_dict.items()]
    header = "|" + "|".join(f"{item}{' ' * (spacing - len(item))}" for item, spacing in zip(table_dict.keys(), max_spacing)) + "|"
    logger.notice(f"Table {table.name()}:")
    logger.notice(" " + "-" * (len(header) - 2) + " ")
    logger.notice(header)
    for i in range(table.rowCount()):
        table_row = "|".join(
            f"{values[i]}{' ' * (spacing - len(str(values[i])))}" for values, spacing in zip(table_dict.values(), max_spacing)
        )
        logger.notice("|" + table_row + "|")
    logger.notice(" " + "-" * (len(header) - 2) + " ")
    return


def create_profiles_table(name, ai):
    table = CreateEmptyTableWorkspace(OutputWorkspace=name)
    table.addColumn(type="str", name="label")
    table.addColumn(type="float", name="mass")
    table.addColumn(type="float", name="intensity")
    table.addColumn(type="float", name="intensity_lb")
    table.addColumn(type="float", name="intensity_ub")
    table.addColumn(type="float", name="width")
    table.addColumn(type="float", name="width_lb")
    table.addColumn(type="float", name="width_ub")
    table.addColumn(type="float", name="center")
    table.addColumn(type="float", name="center_lb")
    table.addColumn(type="float", name="center_ub")

    def wrapb(bound):
        # Literally to just account for NoneType
        if bound is None:
            return np.inf
        return bound

    for mass, intensity, width, center, intensity_bound, width_bound, center_bound in zip(
        ai.masses,
        ai.initial_fitting_parameters[::3],
        ai.initial_fitting_parameters[1::3],
        ai.initial_fitting_parameters[2::3],
        ai.fitting_bounds[::3],
        ai.fitting_bounds[1::3],
        ai.fitting_bounds[2::3],
    ):
        table.addRow(
            [
                str(float(mass)),
                float(mass),
                float(intensity),
                float(wrapb(intensity_bound[0])),
                float(wrapb(intensity_bound[1])),
                float(width),
                float(wrapb(width_bound[0])),
                float(wrapb(width_bound[1])),
                float(center),
                float(wrapb(center_bound[0])),
                float(wrapb(center_bound[1])),
            ]
        )
    return table


def create_table_for_hydrogen_to_mass_ratios():
    table = CreateEmptyTableWorkspace(OutputWorkspace="hydrogen_intensity_ratios_estimates")
    table.addColumn(type="float", name="Hydrogen intensity ratio to chosen mass at each iteration")
    return table


def is_hydrogen_present(masses) -> bool:
    Hmask = np.abs(np.array(masses) - 1) / 1 < 0.1  # H mass whithin 10% of 1 au

    if ~np.any(Hmask):  # H not present
        return False

    print("\nH mass detected.\n")
    assert len(Hmask) > 1, "When H is only mass present, run independent forward procedure, not joint."
    assert Hmask[0], "H mass needs to be the first mass in masses and initPars."
    assert sum(Hmask) == 1, "More than one mass very close to H were detected."
    return True


def ws_history_matches_inputs(runs, mode, ipfile, ws_path):
    if not (ws_path.is_file()):
        logger.notice(f"Cached workspace not found at {ws_path}")
        return False

    ws = Load(Filename=str(ws_path))
    ws_history = ws.getHistory()
    metadata = ws_history.getAlgorithmHistory(0)

    saved_runs = metadata.getPropertyValue("Filename")
    if saved_runs != runs:
        logger.notice(f"Filename in saved workspace did not match: {saved_runs} and {runs}")
        return False

    saved_mode = metadata.getPropertyValue("Mode")
    if saved_mode != mode:
        logger.notice(f"Mode in saved workspace did not match: {saved_mode} and {mode}")
        return False

    saved_ipfile_name = ntpath.basename(metadata.getPropertyValue("InstrumentParFile"))
    if saved_ipfile_name != ipfile:
        logger.notice(f"IP files in saved workspace did not match: {saved_ipfile_name} and {ipfile}")
        return False

    logger.notice("\nLocally saved workspace metadata matched with analysis inputs.\n")
    DeleteWorkspace(ws)
    return True


def save_ws_from_load_vesuvio(runs, mode, ipfile, ws_path):
    if globals.BACKWARD_TAG in ws_path.stem:
        spectra = "3-134"
    elif globals.FORWARD_TAG in ws_path.stem:
        spectra = "135-198"
    else:
        raise ValueError(f"Invalid name to save workspace: {ws_path.name}")

    vesuvio_ws = LoadVesuvio(
        Filename=runs,
        SpectrumList=spectra,
        Mode=mode,
        InstrumentParFile=str(ipfile),
        OutputWorkspace=ws_path.name,
        LoadLogFiles=False,
    )

    SaveNexus(vesuvio_ws, Filename=str(ws_path.absolute()))
    print(f"Workspace saved locally at: {ws_path.absolute()}")
    return


def load_raw_and_empty_from_path(raw_path, empty_path, tof_binning, name, raw_scale_factor, empty_scale_factor, raw_minus_empty):
    print("\nLoading local workspaces ...\n")
    Load(Filename=str(raw_path), OutputWorkspace=name + "_raw")
    Rebin(
        InputWorkspace=name + "_raw",
        Params=tof_binning,
        OutputWorkspace=name + "_raw",
    )
    Scale(
        InputWorkspace=name + "_raw",
        OutputWorkspace=name + "_raw",
        Factor=str(raw_scale_factor),
    )
    SumSpectra(InputWorkspace=name + "_raw", OutputWorkspace=name + "_raw" + "_sum")
    wsToBeFitted = mtd[name + "_raw"]

    if raw_minus_empty:
        Load(Filename=str(empty_path), OutputWorkspace=name + "_empty")
        Rebin(
            InputWorkspace=name + "_empty",
            Params=tof_binning,
            OutputWorkspace=name + "_empty",
        )
        Scale(
            InputWorkspace=name + "_empty",
            OutputWorkspace=name + "_empty",
            Factor=str(empty_scale_factor),
        )
        SumSpectra(InputWorkspace=name + "_empty", OutputWorkspace=name + "_empty" + "_sum")
        wsToBeFitted = Minus(
            LHSWorkspace=name + "_raw",
            RHSWorkspace=name + "_empty",
            OutputWorkspace=name + "_raw_-empty",
        )
    return wsToBeFitted


def cropAndMaskWorkspace(ws, firstSpec, lastSpec, maskedDetectors, maskTOFRange):
    """Returns cloned and cropped workspace with modified name"""
    # Read initial Spectrum number
    wsFirstSpec = ws.getSpectrumNumbers()[0]
    assert firstSpec >= wsFirstSpec, "Can't crop workspace, firstSpec < first spectrum in workspace."

    initialIdx = firstSpec - wsFirstSpec
    lastIdx = lastSpec - wsFirstSpec

    newWsName = ws.name().split("_raw")[0]  # Retrieve original name
    wsCrop = CropWorkspace(
        InputWorkspace=ws,
        StartWorkspaceIndex=initialIdx,
        EndWorkspaceIndex=lastIdx,
        OutputWorkspace=newWsName,
    )

    mask_time_of_flight_bins_with_zeros(wsCrop, maskTOFRange)  # Used to mask resonance peaks

    MaskDetectors(Workspace=wsCrop, SpectraList=maskedDetectors)
    return wsCrop


def mask_time_of_flight_bins_with_zeros(ws, maskTOFRange):
    """
    Masks a given TOF range on ws with zeros on dataY.
    Leaves errors dataE unchanged, as they are used by later treatments.
    Used to mask resonance peaks.
    """

    if maskTOFRange is None:
        return

    dataX, dataY, dataE = extractWS(ws)

    ranges = [r.split("-") for r in maskTOFRange.replace(" ", "").split(",")]
    for r in ranges:
        mask = (dataX >= float(r[0])) & (dataX <= float(r[-1]))
        dataY[mask] = 0

    pass_data_into_ws(dataX, dataY, dataE, ws)
    return


def extractWS(ws):
    """Directly extracts data from workspace into arrays"""
    return ws.extractX(), ws.extractY(), ws.extractE()


def pseudo_voigt(x, sigma, gamma):
    """Convolution between Gaussian with std sigma and Lorentzian with HWHM gamma"""
    fg, fl = 2.0 * sigma * np.sqrt(2.0 * np.log(2.0)), 2.0 * gamma
    f = 0.5346 * fl + np.sqrt(0.2166 * fl**2 + fg**2)
    eta = 1.36603 * fl / f - 0.47719 * (fl / f) ** 2 + 0.11116 * (fl / f) ** 3
    sigma_v, gamma_v = f / (2.0 * np.sqrt(2.0 * np.log(2.0))), f / 2.0
    pseudo_voigt = eta * lorentzian(x, gamma_v) + (1.0 - eta) * gaussian(x, sigma_v)
    return pseudo_voigt


def gaussian(x, sigma):
    """Gaussian centered at zero"""
    gauss = np.exp(-(x**2) / 2 / sigma**2)
    gauss /= np.sqrt(2.0 * np.pi) * sigma
    return gauss


def lorentzian(x, gamma):
    """Lorentzian centered at zero"""
    return gamma / np.pi / (x**2 + gamma**2)


def numerical_third_derivative(x, y):
    k6 = (-y[:, 12:] + y[:, :-12]) * 1
    k5 = (+y[:, 11:-1] - y[:, 1:-11]) * 24
    k4 = (-y[:, 10:-2] + y[:, 2:-10]) * 192
    k3 = (+y[:, 9:-3] - y[:, 3:-9]) * 488
    k2 = (+y[:, 8:-4] - y[:, 4:-8]) * 387
    k1 = (-y[:, 7:-5] + y[:, 5:-7]) * 1584

    dev = k1 + k2 + k3 + k4 + k5 + k6
    dev /= np.power(x[:, 7:-5] - x[:, 6:-6], 3)
    dev /= 12**3
    return dev


def load_resolution(instrument_params):
    """Resolution of parameters to propagate into TOF resolution
    Output: matrix with each parameter in each column"""
    spectra = instrument_params[:, 0]
    L = len(spectra)
    # For spec no below 135, back scattering detectors, mode is double difference
    # For spec no 135 or above, front scattering detectors, mode is single difference
    dE1 = np.where(spectra < 135, 88.7, 73)  # meV, STD
    dE1_lorz = np.where(spectra < 135, 40.3, 24)  # meV, HFHM
    dTOF = np.repeat(0.37, L)  # us
    dTheta = np.repeat(0.016, L)  # rad
    dL0 = np.repeat(0.021, L)  # meters
    dL1 = np.repeat(0.023, L)  # meters

    resolutionPars = np.vstack((dE1, dTOF, dTheta, dL0, dL1, dE1_lorz)).transpose()
    return resolutionPars


def load_instrument_params(ip_file, spectrum_list):
    first_spec = min(spectrum_list)
    last_spec = max(spectrum_list)
    data = np.loadtxt(ip_file, dtype=str)[1:].astype(float)
    spectra = data[:, 0]

    select_rows = np.where((spectra >= first_spec) & (spectra <= last_spec))
    return data[select_rows]


def fix_profile_parameters(incoming_means_table, receiving_profiles_table, h_ratio):
    means_dict = _convert_table_to_dict(incoming_means_table)
    profiles_dict = _convert_table_to_dict(receiving_profiles_table)

    # Set intensities
    for p in profiles_dict.values():
        if np.isclose(p["mass"], 1, atol=0.1):  # Hydrogen present
            p["intensity"] = h_ratio * _get_lightest_profile(means_dict)["mean_intensity"]
            continue
        p["intensity"] = means_dict[p["label"]]["mean_intensity"]

    # Normalise intensities
    sum_intensities = sum([p["intensity"] for p in profiles_dict.values()])
    for p in profiles_dict.values():
        p["intensity"] /= sum_intensities

    # Set widths
    for p in profiles_dict.values():
        try:
            p["width"] = means_dict[p["label"]]["mean_width"]
        except KeyError:
            continue

    # Fix all widths except lightest mass
    for p in profiles_dict.values():
        if p == _get_lightest_profile(profiles_dict):
            continue
        p["width_lb"] = p["width"]
        p["width_ub"] = p["width"]

    result_profiles_table = _convert_dict_to_table(profiles_dict)
    return result_profiles_table


def _convert_table_to_dict(table):
    result_dict = {}
    for i in range(table.rowCount()):
        row_dict = table.row(i)
        result_dict[row_dict["label"]] = row_dict
    return result_dict


def _convert_dict_to_table(m_dict):
    table = CreateEmptyTableWorkspace()
    for p in m_dict.values():
        if table.columnCount() == 0:
            for key, value in p.items():
                value_type = "str" if isinstance(value, str) else "float"
                table.addColumn(value_type, key)

        table.addRow(p)
    return table


def _get_lightest_profile(p_dict):
    profiles = [p for p in p_dict.values()]
    masses = [p["mass"] for p in p_dict.values()]
    return profiles[np.argmin(masses)]


def calculate_h_ratio(means_table, chosen_mass):
    masses = means_table.column("mass")
    intensities = np.array(means_table.column("mean_intensity"))

    if not np.isclose(min(masses), 1, atol=0.1):  # Hydrogen not present
        return None

    # Hydrogen present, assumes its lowest mass
    return intensities[np.argmin(masses)] / intensities[np.argmax(np.isclose(masses, chosen_mass, atol=0.01))]


def extend_range_of_array(arr, n_columns):
    arr = arr.copy()
    left_extend = arr[:, :n_columns] + (arr[:, 0] - arr[:, n_columns]).reshape(-1, 1)
    right_extend = arr[:, -n_columns:] + (arr[:, -1] - arr[:, -n_columns - 1]).reshape(-1, 1)
    return np.concatenate([left_extend, arr, right_extend], axis=-1)


def make_gamma_correction_input_string(masses, mean_widths, mean_intensity_ratios):
    profiles = ""
    for mass, width, intensity in zip(masses, mean_widths, mean_intensity_ratios):
        profiles += "name=GaussianComptonProfile,Mass=" + str(mass) + ",Width=" + str(width) + ",Intensity=" + str(intensity) + ";"
    logger.notice("\nThe sample properties for Gamma Correction are:\n\n" + str(profiles).replace(";", "\n\n").replace(",", "\n"))
    return profiles


def make_multiple_scattering_input_string(masses, meanWidths, meanIntensityRatios):
    atomic_properties_list = np.vstack([masses, meanIntensityRatios, meanWidths]).transpose().flatten().tolist()
    logger.notice(
        "\nSample properties for multiple scattering correction:\n\n"
        + "mass   intensity   width\n"
        + str(np.array(atomic_properties_list).reshape(-1, 3)).replace("[", "").replace("]", "")
        + "\n"
    )
    return atomic_properties_list


def convert_to_list_of_spectrum_numbers(detectors):
    if isinstance(detectors, str):
        detector_ranges = [r.split("-") for r in detectors.replace(" ", "").split(",")]
        return [i for r in detector_ranges for i in range(int(r[0]), int(r[-1]) + 1)]

    if isinstance(detectors, list) or isinstance(detectors, np.ndarray):
        return [int(d) for d in detectors]

    raise ValueError("Type not recognized: Masked detectors should be string, list or array.")



import numpy as np
import matplotlib.pyplot as plt
import scipy
import dill  # Only for converting constraints from string
from pathlib import Path
from mantid.kernel import (
    StringListValidator,
    Direction,
    IntArrayBoundedValidator,
    IntArrayProperty,
    IntBoundedValidator,
    FloatBoundedValidator,
)
from mantid.api import FileProperty, FileAction, PythonAlgorithm, MatrixWorkspaceProperty, WorkspaceGroupProperty
from mantid.dataobjects import TableWorkspaceProperty
from mantid.simpleapi import (
    mtd,
    CreateEmptyTableWorkspace,
    SumSpectra,
    CloneWorkspace,
    DeleteWorkspace,
    VesuvioCalculateGammaBackground,
    VesuvioCalculateMS,
    Scale,
    RenameWorkspace,
    Minus,
    CreateSampleShape,
    VesuvioThickness,
    Integration,
    Divide,
    Multiply,
    DeleteWorkspaces,
    CreateWorkspace,
    GroupWorkspaces,
    SaveAscii,
    SaveNexus,
    AppendSpectra,
)

from mvesuvio.util import handle_config
from mvesuvio.util.analysis_helpers import (
    numerical_third_derivative,
    load_resolution,
    load_instrument_params,
    extend_range_of_array,
    print_table_workspace,
    make_gamma_correction_input_string,
    make_multiple_scattering_input_string,
    pseudo_voigt,
)

try:
    plt.style.use(["ggplot", handle_config.get_plots_config_file()])
except OSError:
    pass
np.set_printoptions(suppress=True, precision=4, linewidth=200)

NEUTRON_MASS = 1.008  # a.m.u.
ENERGY_FINAL = 4906.0  # meV
ENERGY_TO_VELOCITY = 4.3737 * 1.0e-4
VELOCITY_FINAL = np.sqrt(ENERGY_FINAL) * ENERGY_TO_VELOCITY  # m/us
H_BAR = 2.0445


class VesuvioAnalysisRoutine(PythonAlgorithm):
    def summary(self):
        return "Runs the analysis reduction routine for VESUVIO."

    def category(self):
        return "VesuvioAnalysis"

    def PyInit(self):
        self.declareProperty(
            MatrixWorkspaceProperty(name="InputWorkspace", defaultValue="", direction=Direction.Input),
            doc="Workspace to fit Neutron Compton Profiles.",
        )
        self.declareProperty(
            TableWorkspaceProperty(name="InputProfiles", defaultValue="", direction=Direction.Input),
            doc="Table workspace containing starting parameters for profiles.",
        )
        self.declareProperty(
            FileProperty(name="InstrumentParametersFile", defaultValue="", action=FileAction.Load, extensions=["par", "dat"]),
            doc="Filename of the instrument parameter file.",
        )
        self.declareProperty(
            name="ChosenMassIndex",
            defaultValue=0,
            validator=IntBoundedValidator(lower=0),
            doc="Index of mass that anchors hydrogen ratio provided by the user.",
        )
        self.declareProperty(
            name="HRatioToChosenMass",
            defaultValue=0.0,
            validator=FloatBoundedValidator(lower=0),
            doc="Intensity ratio between H peak and chosen mass peak. Traditionally estimated using stoichiometry values.",
        )
        self.declareProperty(name="NumberOfIterations", defaultValue=2, validator=IntBoundedValidator(lower=0))
        self.declareProperty(
            IntArrayProperty(name="InvalidDetectors", validator=IntArrayBoundedValidator(lower=3, upper=198), direction=Direction.Input),
            doc="List of invalid detectors whithin range 3-198.",
        )
        self.declareProperty(name="MultipleScatteringCorrection", defaultValue=False, doc="Whether to run multiple scattering correction.")
        self.declareProperty(name="GammaCorrection", defaultValue=False, doc="Whether to run gamma correction.")
        self.declareProperty(
            name="SampleShapeXml",
            defaultValue="""<cuboid id="sample-shape">
                <left-front-bottom-point x="0.05" y="-0.05" z="0.005" />
                <left-front-top-point x="0.05" y="0.05" z="0.005"/>
                <left-back-bottom-point x="0.05" y="-0.05" z="-0.005" />
                <right-front-bottom-point x="-0.05" y="-0.05" z="0.005" />
                </cuboid>""",
            doc="XML string that describes the shape of the sample. Used in MS correction.",
        )
        self.declareProperty(
            name="ModeRunning",
            defaultValue="BACKWARD",
            validator=StringListValidator(["BACKWARD", "FORWARD"]),
            doc="Whether running backward or forward scattering.",
        )

        self.declareProperty(name="OutputDirectory", defaultValue="", doc="Directory where to save analysis results.")
        self.declareProperty(name="Constraints", defaultValue="", doc="Constraints to use during fitting profiles.")
        self.declareProperty(name="TransmissionGuess", defaultValue=-1.0, validator=FloatBoundedValidator(lower=0, upper=1))
        self.declareProperty(name="MultipleScatteringOrder", defaultValue=-1, validator=IntBoundedValidator(lower=0))
        self.declareProperty(name="NumberOfEvents", defaultValue=-1, validator=IntBoundedValidator(lower=0))
        self.declareProperty(name="ResultsPath", defaultValue="", doc="Directory to store results, to be deleted later.")
        self.declareProperty(name="MinimalOutputFiles", defaultValue=False, doc="Flag to set number of output files to minimum.")
        # Outputs
        self.declareProperty(
            TableWorkspaceProperty(name="OutputMeansTable", defaultValue="", direction=Direction.Output),
            doc="TableWorkspace containing final means of intensity and widths.",
        )
        self.declareProperty(
            WorkspaceGroupProperty(name="OutputNCPGroup", defaultValue="ncp", direction=Direction.Output),
            doc="GroupWorkspace containing Neutron Compton Profiles for each mass and sum.",
        )
        self.declareProperty(
            WorkspaceGroupProperty(name="OutputFSEGroup", defaultValue="fse", direction=Direction.Output),
            doc="GroupWorkspace containing fitted Final State Effects for each mass.",
        )

    def PyExec(self):
        self._setup()
        self.run()

    def _setup(self):
        self._name = self.getPropertyValue("InputWorkspace")
        self._ip_file = self.getProperty("InstrumentParametersFile").value
        self._number_of_iterations = self.getProperty("NumberOfIterations").value
        self._mask_spectra = self.getProperty("InvalidDetectors").value
        self._transmission_guess = self.getProperty("TransmissionGuess").value
        self._multiple_scattering_order = self.getProperty("MultipleScatteringOrder").value
        self._number_of_events = self.getProperty("NumberOfEvents").value
        self._sample_shape_xml = self.getProperty("SampleShapeXml").value
        self._mode_running = self.getProperty("ModeRunning").value
        self._multiple_scattering_correction = self.getProperty("MultipleScatteringCorrection").value
        self._gamma_correction = self.getProperty("GammaCorrection").value
        self._save_results_path = Path(self.getProperty("ResultsPath").value).absolute()
        self._chosen_index_for_h_ratio = self.getProperty("ChosenMassIndex").value
        self._h_ratio = self.getProperty("HRatioToChosenMass").value
        self._constraints = dill.loads(eval(self.getProperty("Constraints").value))
        self._profiles_table = self.getProperty("InputProfiles").value
        self._minimal_output = self.getProperty("MinimalOutputFiles").value

        self._instrument_params = load_instrument_params(self._ip_file, self.getProperty("InputWorkspace").value.getSpectrumNumbers())
        self._resolution_params = load_resolution(self._instrument_params)

        # Create paths if not there already
        self._save_results_path.mkdir(parents=True, exist_ok=True)
        self._save_figures_path = self._save_results_path / "figures"
        self._save_figures_path.mkdir(parents=True, exist_ok=True)

        # Need to transform profiles table into parameter array for optimize.minimize()
        self._initial_fit_parameters = []
        for intensity, width, center in zip(
            self._profiles_table.column("intensity"), self._profiles_table.column("width"), self._profiles_table.column("center")
        ):
            self._initial_fit_parameters.extend([intensity, width, center])

        self._initial_fit_bounds = []

        for intensity_lb, intensity_ub, width_lb, width_ub, center_lb, center_ub in zip(
            self._profiles_table.column("intensity_lb"),
            self._profiles_table.column("intensity_ub"),
            self._profiles_table.column("width_lb"),
            self._profiles_table.column("width_ub"),
            self._profiles_table.column("center_lb"),
            self._profiles_table.column("center_ub"),
        ):
            self._initial_fit_bounds.extend([[intensity_lb, intensity_ub], [width_lb, width_ub], [center_lb, center_ub]])

        # Masses need to be defined in the same order
        self._masses = np.array(self._profiles_table.column("mass"))

        # Variables changing during fit
        self._workspace_for_corrections = self.getProperty("InputWorkspace").value
        self._workspace_being_fit = self.getProperty("InputWorkspace").value
        self._row_being_fit = 0
        self._zero_columns_boolean_mask = None
        self._table_fit_results = None
        self._fit_profiles_workspaces = {}

        assert self._chosen_index_for_h_ratio < self._masses.size, "Index of chosen mass out of range."

    def _update_workspace_data(self):
        self._dataX = self._workspace_being_fit.extractX()
        self._dataY = self._workspace_being_fit.extractY()
        self._dataE = self._workspace_being_fit.extractE()

        self._set_kinematic_arrays(self._dataX)
        self._set_gaussian_resolution()
        self._set_lorentzian_resolution()
        self._set_y_space_arrays()

        self._fit_parameters = np.zeros((len(self._dataY), 3 * self._profiles_table.rowCount() + 2))
        self._row_being_fit = 0
        self._table_fit_results = self._initialize_table_fit_parameters()

        # Initialise workspaces for fitted ncp
        self._fit_profiles_workspaces = {}
        ws_ncp_group = self._initialize_and_group_workspaces("ncp", self._fit_profiles_workspaces)
        self.setPropertyValue("OutputNCPGroup", ws_ncp_group.name())

        # Initialise workspaces for fitted fse
        self._fit_fse_workspaces = {}
        ws_fse_group = self._initialize_and_group_workspaces("fse", self._fit_fse_workspaces)
        self.setPropertyValue("OutputFSEGroup", ws_fse_group.name())

        # Initialise empty means
        self._mean_widths = np.zeros(self._masses.size)
        self._std_widths = np.zeros(self._masses.size)
        self._mean_intensity_ratios = np.zeros(self._masses.size)
        self._std_intensity_ratios = np.zeros(self._masses.size)

    def _initialize_and_group_workspaces(self, suffix, result_dict):
        # Helper to initialize either ncp of fse workspaces and group them together
        assert not result_dict  # Check dict is empty
        for element in self._profiles_table.column("label"):
            result_dict[element] = self._create_emtpy_ncp_workspace(f"_{element}_{suffix}")
        result_dict["total"] = self._create_emtpy_ncp_workspace(f"_total_{suffix}")

        empty_sum_ws = [SumSpectra(ws, OutputWorkspace=ws.name() + "_sum") for ws in result_dict.values()]
        ws_to_group = list(result_dict.values()) + empty_sum_ws
        return GroupWorkspaces(ws_to_group, OutputWorkspace=self._workspace_being_fit.name() + f"_{suffix}_group")

    def _initialize_table_fit_parameters(self):
        table = CreateEmptyTableWorkspace(OutputWorkspace=self._workspace_being_fit.name() + "_fit_results")
        table.setTitle("SciPy Fit Parameters")
        table.addColumn(type="float", name="Spec")
        for label in self._profiles_table.column("label"):
            table.addColumn(type="float", name=f"{label} i")
            table.addColumn(type="float", name=f"{label} w")
            table.addColumn(type="float", name=f"{label} c")
        table.addColumn(type="float", name="NChi2")
        return table

    def _create_emtpy_ncp_workspace(self, suffix):
        return CreateWorkspace(
            DataX=self._dataX,
            DataY=np.zeros(self._dataY.size),
            DataE=np.zeros(self._dataE.size),
            Nspec=self._workspace_being_fit.getNumberHistograms(),
            UnitX="TOF",  # I had hoped for a method like .XUnit() but alas
            OutputWorkspace=self._workspace_being_fit.name() + suffix,
            ParentWorkspace=self._workspace_being_fit,
            Distribution=True,
        )

    def run(self):
        assert self._profiles_table.rowCount() > 0, "Need at least one profile to run the routine!"

        CloneWorkspace(InputWorkspace=self._workspace_being_fit.name(), OutputWorkspace=self._name + "_0")

        for iteration in range(self._number_of_iterations + 1):
            self._workspace_being_fit = mtd[self._name + "_" + str(iteration)]
            self._update_workspace_data()

            self._fit_neutron_compton_profiles()

            self._calculate_summed_workspaces()
            self._save_plots()
            self._set_means_and_std()

            # When last iteration, skip MS and GC
            if iteration == self._number_of_iterations:
                break

            # Do this because MS and Gamma corrections do not accept zero columns
            if iteration == 0:
                self._replace_zeros_with_ncp_for_corrections()

            CloneWorkspace(InputWorkspace=self._workspace_for_corrections.name(), OutputWorkspace="next_iteration")
            self._correct_for_gamma_and_multiple_scattering("next_iteration")

            # Need to remask columns of output of corrections
            self._remask_columns_with_zeros("next_iteration")

            RenameWorkspace(InputWorkspace="next_iteration", OutputWorkspace=self._name + "_" + str(iteration + 1))

        self._save_results()
        return self

    def _fit_neutron_compton_profiles(self):
        """
        Performs the fit of neutron compton profiles to the workspace being fit.
        The profiles are fit on a spectrum by spectrum basis.
        """
        self.log().notice("\nFitting neutron compton profiles ...\n")

        self._row_being_fit = 0
        while self._row_being_fit != len(self._dataY):
            self._fit_neutron_compton_profiles_to_row()
            self._row_being_fit += 1

        assert np.any(self._fit_parameters), "Fitting parameters cannot be zero for all spectra!"
        print_table_workspace(self._table_fit_results)
        return

    def _set_kinematic_arrays(self, dataX):
        # Extend range due to third derivative cutting edges
        dataX = extend_range_of_array(dataX, 6)

        det, plick, angle, T0, L0, L1 = np.hsplit(self._instrument_params, 6)  # each is of len(dataX)

        # T0 is electronic delay due to instruments
        t_us = dataX - T0
        self._v0 = VELOCITY_FINAL * L0 / (VELOCITY_FINAL * t_us - L1)
        # en_to_vel is a factor used to easily change velocity to energy and vice-versa
        self._E0 = np.square(self._v0 / ENERGY_TO_VELOCITY)
        self._deltaE = self._E0 - ENERGY_FINAL
        delta_Q2 = (
            2.0
            * NEUTRON_MASS
            / H_BAR**2
            * (self._E0 + ENERGY_FINAL - 2.0 * np.sqrt(self._E0 * ENERGY_FINAL) * np.cos(angle / 180.0 * np.pi))
        )
        self._deltaQ = np.sqrt(delta_Q2)
        return

    def _set_y_space_arrays(self):
        delta_Q = self._deltaQ[np.newaxis, :, :]
        delta_E = self._deltaE[np.newaxis, :, :]
        masses = self._masses.reshape(-1, 1, 1)

        energy_recoil = np.square(H_BAR * delta_Q) / 2.0 / masses
        y_spaces = masses / H_BAR**2 / delta_Q * (delta_E - energy_recoil)

        # Swap axis so that first axis selects spectra
        self._y_space_arrays = np.swapaxes(y_spaces, 0, 1)
        return

    def _save_plots(self):
        if self._minimal_output or not self._save_figures_path.exists():
            return

        fig, ax = plt.subplots(subplot_kw={"projection": "mantid"})

        ws_data_sum = mtd[self._workspace_being_fit.name() + "_sum"]
        lab = "Sum of spectra"
        ax.errorbar(ws_data_sum, fmt="k.", label="Sum of spectra", elinewidth=1.5)

        label_list = [lab]
        ws_fig_name = self._workspace_being_fit.name() + "_ncp_fits_sum"
        CloneWorkspace(ws_data_sum, OutputWorkspace=ws_fig_name)

        for key, ws in self._fit_profiles_workspaces.items():
            ws_sum = mtd[ws.name() + "_sum"]
            lab = f"Sum of {key} profile"
            ax.plot(ws_sum, label=lab)
            label_list.append(lab)
            AppendSpectra(ws_fig_name, ws_sum, OutputWorkspace=ws_fig_name)

        ax.set_xlabel("TOF")
        ax.set_ylabel("Counts")
        ax.set_title("Sum of NCP fits")
        ax.legend()

        fileName = self._workspace_being_fit.name() + "_profiles_sum.pdf"
        savePath = self._save_figures_path / fileName
        plt.savefig(savePath, bbox_inches="tight")
        plt.close(fig)

        # TODO: Sort this out into something cleaner
        file_name = self._workspace_being_fit.name() + "_profiles_sum.txt"
        SaveAscii(ws_fig_name, str(self._save_results_path / file_name), LogList=label_list)
        return

    def _calculate_summed_workspaces(self):
        SumSpectra(InputWorkspace=self._workspace_being_fit.name(), OutputWorkspace=self._workspace_being_fit.name() + "_sum")

        for ws in self._fit_profiles_workspaces.values():
            SumSpectra(InputWorkspace=ws.name(), OutputWorkspace=ws.name() + "_sum")

        for ws in self._fit_fse_workspaces.values():
            SumSpectra(InputWorkspace=ws.name(), OutputWorkspace=ws.name() + "_sum")

    def _set_means_and_std(self):
        widths = np.zeros((self._profiles_table.rowCount(), self._table_fit_results.rowCount()))
        intensities = np.zeros(widths.shape)

        for i, label in enumerate(self._profiles_table.column("label")):
            widths[i] = self._table_fit_results.column(f"{label} w")
            intensities[i] = self._table_fit_results.column(f"{label} i")
            self._set_means_and_std_arrays(widths, intensities)

        self._create_means_table()
        return

    def _set_means_and_std_arrays(self, widths, intensities):
        # Remove failed fits and masked spectra
        non_zero_columns = np.any(widths != 0, axis=0)
        widths = widths[:, non_zero_columns]
        intensities = intensities[:, non_zero_columns]

        widths_mean = np.mean(widths, axis=1).reshape(-1, 1)
        widths_std = np.std(widths, axis=1).reshape(-1, 1)

        widths_deviations = np.abs(widths - widths_mean)

        # Remove width outliers
        widths[widths_deviations > widths_std] = np.nan
        intensities[widths_deviations > widths_std] = np.nan

        # Use sum instead of nansum to propagate nans
        intensities = intensities / intensities.sum(axis=0)

        self._mean_widths = np.nanmean(widths, axis=1)
        self._std_widths = np.nanstd(widths, axis=1)
        self._mean_intensity_ratios = np.nanmean(intensities, axis=1)
        self._std_intensity_ratios = np.nanstd(intensities, axis=1)
        return

    def _create_means_table(self):
        table = CreateEmptyTableWorkspace(OutputWorkspace=self._workspace_being_fit.name() + "_means")
        table.addColumn(type="str", name="label")
        table.addColumn(type="float", name="mass")
        table.addColumn(type="float", name="mean_width")
        table.addColumn(type="float", name="std_width")
        table.addColumn(type="float", name="mean_intensity")
        table.addColumn(type="float", name="std_intensity")

        for label, mass, mean_width, std_width, mean_intensity, std_intensity in zip(
            self._profiles_table.column("label"),
            self._masses,
            self._mean_widths,
            self._std_widths,
            self._mean_intensity_ratios,
            self._std_intensity_ratios,
        ):
            # Explicit conversion to float required to match profiles table
            table.addRow([label, float(mass), float(mean_width), float(std_width), float(mean_intensity), float(std_intensity)])

        self.setPropertyValue("OutputMeansTable", table.name())
        print_table_workspace(table, precision=5)
        return table

    def _fit_neutron_compton_profiles_to_row(self):
        if np.all(self._dataY[self._row_being_fit] == 0):
            self._table_fit_results.addRow(np.zeros(3 * self._profiles_table.rowCount() + 2))
            spectrum_number = self._instrument_params[self._row_being_fit, 0]
            self.log().notice(f"Skip spectrum {int(spectrum_number):3d}")
            return

        result = scipy.optimize.minimize(
            self._error_function,
            self._initial_fit_parameters,
            method="SLSQP",
            bounds=self._initial_fit_bounds,
            constraints=self._constraints,
            tol=1e-6,
        )
        fitPars = result["x"]

        # Pass fit parameters to results table
        noDegreesOfFreedom = len(self._dataY[self._row_being_fit]) - len(fitPars)
        normalised_chi2 = result["fun"] / noDegreesOfFreedom
        spectrum_number = self._instrument_params[self._row_being_fit, 0]
        tableRow = np.hstack((spectrum_number, fitPars, normalised_chi2))
        self._table_fit_results.addRow(tableRow)

        # Store results for easier access when calculating means
        self._fit_parameters[self._row_being_fit] = tableRow

        self.log().notice(f"Fit spectrum {int(spectrum_number):3d}: \u2713")

        # Pass fit profiles into workspaces
        ncp_for_each_mass, fse_for_each_mass = self._neutron_compton_profiles(fitPars)
        for ncp, fse, element in zip(ncp_for_each_mass, fse_for_each_mass, self._profiles_table.column("label")):
            self._fit_profiles_workspaces[element].dataY(self._row_being_fit)[:] = ncp
            self._fit_fse_workspaces[element].dataY(self._row_being_fit)[:] = fse

        self._fit_profiles_workspaces["total"].dataY(self._row_being_fit)[:] = np.sum(ncp_for_each_mass, axis=0)
        self._fit_fse_workspaces["total"].dataY(self._row_being_fit)[:] = np.sum(fse_for_each_mass, axis=0)
        return

    def _error_function(self, pars):
        """Error function to be minimized, in TOF space"""

        ncp_for_each_mass, fse_for_each_mass = self._neutron_compton_profiles(pars)

        ncp_total = np.sum(ncp_for_each_mass, axis=0)
        data_y = self._dataY[self._row_being_fit]
        data_e = self._dataE[self._row_being_fit]

        # Ignore any masked values on tof range
        nonzero_mask = np.nonzero(data_y)
        ncp_total = ncp_total[nonzero_mask]
        data_y = data_y[nonzero_mask]
        data_e = data_e[nonzero_mask]

        if np.all(data_e == 0):  # When errors not present
            return np.sum((ncp_total - data_y) ** 2)

        return np.sum((ncp_total - data_y) ** 2 / data_e**2)

    def _neutron_compton_profiles(self, pars):
        """
        Neutron Compton Profile distribution on TOF space for a single spectrum.
        Calculated from kinematics, J(y) and resolution functions.
        """
        intensities = pars[::3].reshape(-1, 1)
        widths = pars[1::3].reshape(-1, 1)
        centers = pars[2::3].reshape(-1, 1)
        masses = self._masses.reshape(-1, 1)

        gaussian_width = self._get_gaussian_resolution(centers)
        lorentzian_width = self._get_lorentzian_resolution(centers)
        total_gaussian_width = np.sqrt(widths**2 + gaussian_width**2)

        JOfY = pseudo_voigt(self._y_space_arrays[self._row_being_fit] - centers, total_gaussian_width, lorentzian_width)

        # Third derivative cuts edges of array by 6 indices
        JOfY_third_derivative = numerical_third_derivative(self._y_space_arrays[self._row_being_fit], JOfY)

        deltaQ = self._deltaQ[self._row_being_fit, 6:-6]
        E0 = self._E0[self._row_being_fit, 6:-6]
        JOfY = JOfY[:, 6:-6]

        FSE = -JOfY_third_derivative * widths**4 / deltaQ * 0.72

        NCP = intensities * (JOfY + FSE) * E0 * E0 ** (-0.92) * masses / deltaQ
        FSE = intensities * FSE * E0 * E0 ** (-0.92) * masses / deltaQ
        return NCP, FSE

    def _get_gaussian_resolution(self, centers):
        proximity_to_y_centers = np.abs(self._y_space_arrays[self._row_being_fit] - centers)
        gauss_resolution = self._gaussian_resolution[self._row_being_fit]
        assert proximity_to_y_centers.shape == gauss_resolution.shape
        return np.take_along_axis(gauss_resolution, proximity_to_y_centers.argmin(axis=1, keepdims=True), axis=1)

    def _set_gaussian_resolution(self):
        masses = self._masses.reshape(-1, 1, 1)
        v0 = self._v0
        E0 = self._E0
        delta_Q = self._deltaQ
        det, plick, angle, T0, L0, L1 = np.hsplit(self._instrument_params, 6)
        dE1, dTOF, dTheta, dL0, dL1, dE1_lorz = np.hsplit(self._resolution_params, 6)

        angle = angle * np.pi / 180

        dWdE1 = 1.0 + (E0 / ENERGY_FINAL) ** 1.5 * (L1 / L0)
        dWdTOF = 2.0 * E0 * v0 / L0
        dWdL1 = 2.0 * E0**1.5 / ENERGY_FINAL**0.5 / L0
        dWdL0 = 2.0 * E0 / L0

        dW2 = (dWdE1**2 * dE1**2 + dWdTOF**2 * dTOF**2 + dWdL1**2 * dL1**2 + dWdL0**2 * dL0**2) * np.ones((masses.size, 1, 1))
        # conversion from meV^2 to A^-2, dydW = (M/q)^2
        dW2 *= (masses / H_BAR**2 / delta_Q) ** 2

        dQdE1 = 1.0 - (E0 / ENERGY_FINAL) ** 1.5 * L1 / L0 - np.cos(angle) * ((E0 / ENERGY_FINAL) ** 0.5 - L1 / L0 * E0 / ENERGY_FINAL)
        dQdTOF = 2.0 * E0 * v0 / L0
        dQdL1 = 2.0 * E0**1.5 / L0 / ENERGY_FINAL**0.5
        dQdL0 = 2.0 * E0 / L0
        dQdTheta = 2.0 * np.sqrt(E0 * ENERGY_FINAL) * np.sin(angle)

        dQ2 = (
            dQdE1**2 * dE1**2
            + (dQdTOF**2 * dTOF**2 + dQdL1**2 * dL1**2 + dQdL0**2 * dL0**2) * np.abs(ENERGY_FINAL / E0 * np.cos(angle) - 1)
            + dQdTheta**2 * dTheta**2
        )
        dQ2 *= (NEUTRON_MASS / H_BAR**2 / delta_Q) ** 2

        # in A-1    #same as dy^2 = (dy/dw)^2*dw^2 + (dy/dq)^2*dq^2
        gaussianResWidth = np.sqrt(dW2 + dQ2)
        self._gaussian_resolution = np.swapaxes(gaussianResWidth, 0, 1)
        return

    def _get_lorentzian_resolution(self, centers):
        proximity_to_y_centers = np.abs(self._y_space_arrays[self._row_being_fit] - centers)
        lorentzian_resolution = self._lorentzian_resolution[self._row_being_fit]
        assert proximity_to_y_centers.shape == lorentzian_resolution.shape
        return np.take_along_axis(lorentzian_resolution, proximity_to_y_centers.argmin(axis=1, keepdims=True), axis=1)

    def _set_lorentzian_resolution(self):
        masses = self._masses.reshape(-1, 1, 1)
        E0 = self._E0
        delta_Q = self._deltaQ
        det, plick, angle, T0, L0, L1 = np.hsplit(self._instrument_params, 6)
        dE1, dTOF, dTheta, dL0, dL1, dE1_lorz = np.hsplit(self._resolution_params, 6)

        angle = angle * np.pi / 180

        dWdE1_lor = (1.0 + (E0 / ENERGY_FINAL) ** 1.5 * (L1 / L0)) ** 2 * np.ones((masses.size, 1, 1))
        # conversion from meV^2 to A^-2
        dWdE1_lor *= (masses / H_BAR**2 / delta_Q) ** 2

        dQdE1_lor = (
            1.0 - (E0 / ENERGY_FINAL) ** 1.5 * L1 / L0 - np.cos(angle) * ((E0 / ENERGY_FINAL) ** 0.5 + L1 / L0 * E0 / ENERGY_FINAL)
        ) ** 2
        dQdE1_lor *= (NEUTRON_MASS / H_BAR**2 / delta_Q) ** 2

        lorentzianResWidth = np.sqrt(dWdE1_lor + dQdE1_lor) * dE1_lorz  # in A-1
        self._lorentzian_resolution = np.swapaxes(lorentzianResWidth, 0, 1)
        return

    def _replace_zeros_with_ncp_for_corrections(self):
        """
        If the initial input contains columns with zeros
        (to mask resonance peaks) then these sections must be approximated
        by the total fitted function because multiple scattering and
        gamma correction algorithms do not accept columns with zeros.
        If no masked columns are present then the input workspace
        for corrections is left unchanged.
        """
        dataY = self._workspace_for_corrections.extractY()
        ncp = self._fit_profiles_workspaces["total"].extractY()

        self._zero_columns_boolean_mask = np.all(dataY == 0, axis=0)  # Masked Cols

        for row in range(self._workspace_for_corrections.getNumberHistograms()):
            self._workspace_for_corrections.dataY(row)[self._zero_columns_boolean_mask] = ncp[row, self._zero_columns_boolean_mask]

        SumSpectra(InputWorkspace=self._workspace_for_corrections.name(), OutputWorkspace=self._workspace_for_corrections.name() + "_sum")
        return

    def _remask_columns_with_zeros(self, ws_to_remask_name):
        """
        Uses previously stored information on masked columns in the
        initial workspace to set these columns again to zero on the
        workspace resulting from the multiple scattering or gamma correction.
        """
        ws_to_remask = mtd[ws_to_remask_name]
        for row in range(ws_to_remask.getNumberHistograms()):
            ws_to_remask.dataY(row)[self._zero_columns_boolean_mask] = 0
            ws_to_remask.dataE(row)[self._zero_columns_boolean_mask] = 0
        return

    def _correct_for_gamma_and_multiple_scattering(self, ws_name):
        if self._gamma_correction:
            gamma_correction_ws = self.create_gamma_workspaces()
            Minus(LHSWorkspace=ws_name, RHSWorkspace=gamma_correction_ws.name(), OutputWorkspace=ws_name)

        if self._multiple_scattering_correction:
            multiple_scattering_ws = self.create_multiple_scattering_workspaces()
            Minus(LHSWorkspace=ws_name, RHSWorkspace=multiple_scattering_ws.name(), OutputWorkspace=ws_name)
        return

    def create_multiple_scattering_workspaces(self):
        """Creates _MulScattering and _TotScattering workspaces used for the MS correction"""
        self.log().notice("\nEvaluating multiple scattering correction ...\n")

        CreateSampleShape(self._workspace_for_corrections, self._sample_shape_xml)

        # Make local variables
        masses = self._masses
        mean_widths = self._mean_widths
        mean_intensity_ratios = self._mean_intensity_ratios

        # If Backscattering mode and H is present in the sample, add H to MS properties
        if self._mode_running == "BACKWARD":
            if self._h_ratio > 0:  # If H is present, ratio is a number
                HIntensity = self._h_ratio * mean_intensity_ratios[self._chosen_index_for_h_ratio]
                mean_intensity_ratios = np.append(mean_intensity_ratios, HIntensity)
                mean_intensity_ratios /= np.sum(mean_intensity_ratios)
                masses = np.append(masses, 1.0079)
                mean_widths = np.append(mean_widths, 5.0)

        dens, trans = VesuvioThickness(
            Masses=masses,
            Amplitudes=mean_intensity_ratios,
            TransmissionGuess=self._transmission_guess,
            Thickness=0.1,
        )
        ws_corrections_name = self._workspace_for_corrections.name()
        atomic_properties_list = make_multiple_scattering_input_string(masses, mean_widths, mean_intensity_ratios)
        VesuvioCalculateMS(
            InputWorkspace=ws_corrections_name,
            NoOfMasses=len(masses),
            SampleDensity=dens.cell(9, 1),
            AtomicProperties=atomic_properties_list,
            BeamRadius=2.5,
            NumScatters=self._multiple_scattering_order,
            NumEventsPerRun=int(self._number_of_events),
            TotalScatteringWS=ws_corrections_name + "_tot_sctr",
            MultipleScatteringWS=ws_corrections_name + "_mltp_sctr",
        )
        data_normalisation = Integration(ws_corrections_name)
        simulation_normalisation = Integration(ws_corrections_name + "_tot_sctr")
        for ws_sctr_name in (ws_corrections_name + "_mltp_sctr", ws_corrections_name + "_tot_sctr"):
            Divide(
                LHSWorkspace=ws_sctr_name,
                RHSWorkspace=simulation_normalisation,
                OutputWorkspace=ws_sctr_name,
            )
            Multiply(
                LHSWorkspace=ws_sctr_name,
                RHSWorkspace=data_normalisation,
                OutputWorkspace=ws_sctr_name,
            )
            # Sum spectra for vizualisation
            SumSpectra(ws_sctr_name, OutputWorkspace=ws_sctr_name + "_sum")

        DeleteWorkspaces([data_normalisation, simulation_normalisation, trans, dens])
        # The only remaining workspaces are the _mltp_sctr and _tot_sctr
        return mtd[ws_corrections_name + "_mltp_sctr"]

    def create_gamma_workspaces(self):
        """Creates _gamma_background correction workspace to be subtracted from the main workspace"""

        ws_corrections_name = self._workspace_for_corrections.name()
        profiles_string = make_gamma_correction_input_string(self._masses, self._mean_widths, self._mean_intensity_ratios)

        background, corrected = VesuvioCalculateGammaBackground(InputWorkspace=ws_corrections_name, ComptonFunction=profiles_string)
        DeleteWorkspace(corrected)
        RenameWorkspace(InputWorkspace=background, OutputWorkspace=ws_corrections_name + "_gamma_backgr")
        Scale(
            InputWorkspace=ws_corrections_name + "_gamma_backgr",
            OutputWorkspace=ws_corrections_name + "_gamma_backgr",
            Factor=0.9,
            Operation="Multiply",
        )
        return mtd[ws_corrections_name + "_gamma_backgr"]

    def _save_results(self):
        all_workspaces = mtd.getObjectNames()
        if self._minimal_output:
            last_iteration = max([ws.replace("_total_ncp", "")[-1] for ws in all_workspaces if ws.endswith("_total_ncp")])
            for ws_name in all_workspaces:
                if ws_name.endswith((f"{last_iteration}_total_ncp", f"{last_iteration}_total_fse")):
                    SaveAscii(ws_name, str(self._save_results_path / ws_name) + ".txt")
                if ws_name.endswith((f"{last_iteration}_fit_results", f"{last_iteration}_means")):
                    SaveAscii(ws_name, str(self._save_results_path / ws_name) + ".txt")
            return

        for ws_name in all_workspaces:
            if ws_name.endswith(("ncp", "fse", "initial_parameters", "means", "fit_results")):
                SaveAscii(ws_name, str(self._save_results_path / ws_name) + ".txt")
            if ws_name.endswith(tuple([str(i) for i in range(10)])):
                SaveNexus(ws_name, str(self._save_results_path / ws_name) + ".nxs")



"""Package defining entry points."""

import argparse
from os import path
from pathlib import Path
from mvesuvio.util import handle_config


def main(manual_args=None):
    args = _setup_and_parse_args() if not manual_args else manual_args
    if args.command == "version":
        _print_version()

    if args.command == "config":
        _setup_config(args)

    if args.command == "run":
        _setup_config(None)
        _run_analysis(args)

    if args.command == "bootstrap":
        _setup_config(None)
        _run_bootstrap(args)


def _setup_and_parse_args():
    parser = _set_up_parser()
    args = parser.parse_args()
    return args


def _set_up_parser():
    parser = argparse.ArgumentParser(description="Package to analyse Vesuvio instrument data")
    subparsers = parser.add_subparsers(dest="command", required=True)
    subparsers.add_parser("version", help="Display the version of mvesuvio")
    config_parser = subparsers.add_parser("config", help="Set mvesuvio configuration")
    config_parser.add_argument("--analysis-inputs", "-i", help="Set the inputs python file", default="", type=str)
    config_parser.add_argument(
        "--ip-folder",
        "-p",
        help="Set the intrument parameters directory",
        default="",
        type=str,
    )

    run_parser = subparsers.add_parser("run", help="Run mvesuvio analysis")
    run_parser.add_argument(
        "--back-workspace",
        "-b",
        help="Input workspace for vesuvio backward analysis, bypasses loading (and subtracting) raw and empty.",
        default="",
        type=str,
    )
    run_parser.add_argument(
        "--front-workspace",
        "-f",
        help="Input workspace for vesuvio forward analysis, bypasses loading (and subtracting) raw and empty.",
        default="",
        type=str,
    )
    run_parser.add_argument(
        "--minimal-output",
        action="store_true",
        help="Flag to set output files to minimum.",
        default=False,
    )
    run_parser.add_argument(
        "--outputs-dir",
        "-o",
        help="Directory for populating with output files.",
        default="",
        type=str,
    )
    boot_parser = subparsers.add_parser("bootstrap", help="Run bootstrap of vesuvio analysis (without y-space fitting)")
    boot_parser.add_argument(
        "--inputs-dir",
        "-d",
        help="Directory containing input bootstrap replicas. Replicas should be inside sparate backaward and forward subdirectories.",
        default="",
        type=str,
    )
    return parser


def _setup_config(args):
    __set_logging_properties()

    handle_config.refresh_config_dir_and_contents()

    if not handle_config.is_cache_set():
        handle_config.set_default_config_vars()

    inputs = handle_config.read_cached_var("caching.inputs")
    ipfolder_dir = handle_config.read_cached_var("caching.ipfolder")

    if args and args.analysis_inputs:
        inputs = str(Path(args.analysis_inputs).absolute())
    if args and args.ip_folder:
        ipfolder_dir = str(Path(args.ip_folder).absolute())

    handle_config.set_config_vars(
        {
            "caching.inputs": inputs,
            "caching.ipfolder": ipfolder_dir,
        }
    )
    handle_config.check_dir_exists("IP folder", ipfolder_dir)


def __set_logging_properties():
    from mantid.kernel import ConfigService

    ConfigService.setString("logging.loggers.root.channel.class", "SplitterChannel")
    ConfigService.setString("logging.loggers.root.channel.channel1", "consoleChannel")
    ConfigService.setString("logging.loggers.root.channel.channel2", "fileChannel")
    ConfigService.setString("logging.channels.consoleChannel.class", "ConsoleChannel")
    ConfigService.setString("logging.channels.fileChannel.class", "FileChannel")
    ConfigService.setString("logging.channels.fileChannel.path", "mantid.log")
    ConfigService.setString("logging.channels.fileChannel.formatter.class", "PatternFormatter")
    ConfigService.setString("logging.channels.fileChannel.formatter.pattern", "%Y-%m-%d %H:%M:%S,%i [%I] %p %s - %t")
    # Set properties on Mantid.user.properties not working due to Mantid bug
    # Need to set properties on file in Mantid installation
    mantid_properties_file = path.join(ConfigService.getPropertiesDir(), "Mantid.properties")
    ConfigService.saveConfig(mantid_properties_file)
    return


def _run_analysis(args):
    from mvesuvio.main.run_routine import Runner

    if not args:
        Runner().run()
        return
    Runner(
        override_back_workspace=args.back_workspace,
        override_front_workspace=args.front_workspace,
        minimal_output=args.minimal_output,
        output_directory=args.outputs_dir,
    ).run()


def _run_bootstrap(args):
    from mvesuvio.main.run_routine import Runner

    if not args:
        return
    Runner(bootstrap_inputs_directory=args.inputs_dir, minimal_output=True).run_bootstrap()


def _print_version():
    from mvesuvio import version

    print(version())


if __name__ == "__main__":
    main()



import os
from shutil import copyfile, copytree, ignore_patterns
from pathlib import Path


### PATH CONSTANTS ###
VESUVIO_PACKAGE_PATH = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
VESUVIO_CONFIG_PATH = os.path.join(os.path.expanduser("~"), ".mvesuvio")
VESUVIO_PROPERTIES_FILE = "vesuvio.user.properties"
ANALYSIS_INPUTS_FILE = "analysis_inputs.py"
MANTID_CONFIG_FILE = "Mantid.user.properties"
PLOTS_CONFIG_FILE = "vesuvio.plots.mplstyle"
SCRIPT_FIGUES_FILE = "script_to_create_figures.py"
IP_FOLDER = "ip_files"
######################


def set_default_config_vars():
    set_config_vars(
        {
            "caching.inputs": str(Path(VESUVIO_CONFIG_PATH, ANALYSIS_INPUTS_FILE)),
            "caching.ipfolder": str(Path(VESUVIO_CONFIG_PATH, IP_FOLDER)),
        }
    )


def __read_config(config_file_path, throw_on_not_found=True):
    lines = ""
    try:
        with open(config_file_path, "r") as file:
            lines = file.readlines()
    except IOError:
        if throw_on_not_found:
            raise RuntimeError(f"Could not read from vesuvio config file: {config_file_path}")
    return lines


def set_config_vars(var_dict):
    file_path = Path(VESUVIO_PACKAGE_PATH, "config", VESUVIO_PROPERTIES_FILE)
    lines = __read_config(file_path)

    updated_lines = []
    for line in lines:
        match = False
        for var in var_dict:
            if line.startswith(var):
                new_line = f"{var}={var_dict[var]}"
                updated_lines.append(f"{new_line}\n")
                match = True
                print(f"Setting: {new_line}")
                break

        if not match:
            updated_lines.append(line)

    with open(file_path, "w") as file:
        file.writelines(updated_lines)


def read_cached_var(var, throw_on_not_found=True):
    file_path = Path(VESUVIO_PACKAGE_PATH, "config", VESUVIO_PROPERTIES_FILE)
    lines = __read_config(file_path, throw_on_not_found)

    result = ""
    for line in lines:
        if line.startswith(var):
            result = line.split("=", 2)[1].strip("\n")
            break
    if not result and throw_on_not_found:
        raise ValueError(f"{var} was not found in the vesuvio config")
    return result


def get_script_name():
    filename = os.path.basename(read_cached_var("caching.inputs"))
    scriptName = filename.removesuffix(".py")
    return scriptName


def get_plots_config_file() -> str:
    return os.path.abspath(os.path.join(VESUVIO_CONFIG_PATH, PLOTS_CONFIG_FILE))


def refresh_config_dir_and_contents():
    setup_config_dir()
    setup_plots_config_file()
    setup_script_figures_file()
    setup_default_inputs()
    setup_default_ipfile_dir()


def setup_script_figures_file():
    if not Path(VESUVIO_CONFIG_PATH, SCRIPT_FIGUES_FILE).exists():
        copyfile(
            Path(VESUVIO_PACKAGE_PATH, "config", SCRIPT_FIGUES_FILE),
            Path(VESUVIO_CONFIG_PATH, SCRIPT_FIGUES_FILE),
        )


def setup_plots_config_file():
    if not Path(VESUVIO_CONFIG_PATH, PLOTS_CONFIG_FILE).exists():
        copyfile(
            Path(VESUVIO_PACKAGE_PATH, "config", PLOTS_CONFIG_FILE),
            Path(VESUVIO_CONFIG_PATH, PLOTS_CONFIG_FILE),
        )


def setup_config_dir():
    if not os.path.isdir(VESUVIO_CONFIG_PATH):
        os.makedirs(VESUVIO_CONFIG_PATH)


def setup_default_inputs():
    copyfile(
        Path(VESUVIO_PACKAGE_PATH, "config", ANALYSIS_INPUTS_FILE),
        Path(VESUVIO_CONFIG_PATH, ANALYSIS_INPUTS_FILE),
    )


def setup_default_ipfile_dir():
    if not Path(VESUVIO_CONFIG_PATH, IP_FOLDER).is_dir():
        copytree(
            Path(VESUVIO_PACKAGE_PATH, "config", "ip_files"),
            Path(VESUVIO_CONFIG_PATH, IP_FOLDER),
            ignore=ignore_patterns("__*"),
        )


def is_cache_set():
    if read_cached_var("caching.inputs", False):
        return True
    else:
        return False


def check_dir_exists(type, path):
    if not os.path.isdir(path):
        print(f"Directory of {type} could not be found at location: {path}")
        return False
    return True
